@@ -83,8 +83,10 @@ struct option {
 
 #include "bsdtar.h"
 #include "crypto.h"
+#include "humansize.h"
 #include "network.h"
 #include "sysendian.h"
+#include "tarsnap_opt.h"
 
 #if !HAVE_DECL_OPTARG
 extern int optarg;
@@ -94,6 +96,12 @@ extern int optarg;
 extern int optind;
 #endif
 
+/* Global tarsnap options declared in tarsnap_opt.h. */
+int tarsnap_opt_aggressive_networking = 0;
+int tarsnap_opt_humanize_numbers = 0;
+int tarsnap_opt_noisy_warnings = 0;
+uint64_t tarsnap_opt_maxbytesout = (uint64_t)(-1);
+
 /* External function to parse a date/time string (from getdate.y) */
 time_t get_date(const char *);
 
@@ -135,17 +143,20 @@ static const char *tar_opts = "+@:BC:cdf:HhI:kLlmnOoPprtT:UvW:wX:x";
 
 /* Fake short equivalents for long options that otherwise lack them. */
 enum {
-	OPTION_CACHEDIR=1,
+	OPTION_AGGRESSIVE_NETWORKING=1,
+	OPTION_CACHEDIR,
 	OPTION_CHECK_LINKS,
 	OPTION_CHROOT,
 	OPTION_EXCLUDE,
 	OPTION_FSCK,
 	OPTION_HELP,
 	OPTION_INCLUDE,
+	OPTION_HUMANIZE_NUMBERS,
 	OPTION_KEYFILE,
 	OPTION_KEEP_NEWER_FILES,
 	OPTION_LIST_ARCHIVES,
 	OPTION_LOWMEM,
+	OPTION_MAXBW,
 	OPTION_NEWER_CTIME,
 	OPTION_NEWER_CTIME_THAN,
 	OPTION_NEWER_MTIME,
@@ -153,6 +164,8 @@ enum {
 	OPTION_NODUMP,
 	OPTION_NO_SAME_OWNER,
 	OPTION_NO_SAME_PERMISSIONS,
+	OPTION_NOISY_WARNINGS,
+	OPTION_NUKE,
 	OPTION_NULL,
 	OPTION_ONE_FILE_SYSTEM,
 	OPTION_PRINT_STATS,
@@ -170,6 +183,7 @@ enum {
  */
 static const struct option tar_longopts[] = {
 	{ "absolute-paths",     no_argument,       NULL, 'P' },
+	{ "aggressive-networking", no_argument,	   NULL, OPTION_AGGRESSIVE_NETWORKING },
 	{ "cachedir",		required_argument, NULL, OPTION_CACHEDIR },
 	{ "cd",                 required_argument, NULL, 'C' },
 	{ "check-links",        no_argument,       NULL, OPTION_CHECK_LINKS },
@@ -186,6 +200,7 @@ static const struct option tar_longopts[] = {
 	{ "files-from",         required_argument, NULL, 'T' },
 	{ "fsck",		no_argument,	   NULL, OPTION_FSCK },
 	{ "help",               no_argument,       NULL, OPTION_HELP },
+	{ "humanize-numbers",	no_argument,	   NULL, OPTION_HUMANIZE_NUMBERS },
 	{ "include",            required_argument, NULL, OPTION_INCLUDE },
 	{ "interactive",        no_argument,       NULL, 'w' },
 	{ "insecure",           no_argument,       NULL, 'P' },
@@ -195,6 +210,7 @@ static const struct option tar_longopts[] = {
 	{ "list",               no_argument,       NULL, 't' },
 	{ "list-archives",	no_argument,	   NULL, OPTION_LIST_ARCHIVES },
 	{ "lowmem",		no_argument,	   NULL, OPTION_LOWMEM },
+	{ "maxbw",		required_argument, NULL, OPTION_MAXBW },
 	{ "modification-time",  no_argument,       NULL, 'm' },
 	{ "newer",		required_argument, NULL, OPTION_NEWER_CTIME },
 	{ "newer-ctime",	required_argument, NULL, OPTION_NEWER_CTIME },
@@ -203,10 +219,12 @@ static const struct option tar_longopts[] = {
 	{ "newer-mtime-than",	required_argument, NULL, OPTION_NEWER_MTIME_THAN },
 	{ "newer-than",		required_argument, NULL, OPTION_NEWER_CTIME_THAN },
 	{ "nodump",             no_argument,       NULL, OPTION_NODUMP },
+	{ "noisy-warnings",	no_argument,	   NULL, OPTION_NOISY_WARNINGS },
 	{ "norecurse",          no_argument,       NULL, 'n' },
 	{ "no-recursion",       no_argument,       NULL, 'n' },
 	{ "no-same-owner",	no_argument,	   NULL, OPTION_NO_SAME_OWNER },
 	{ "no-same-permissions",no_argument,	   NULL, OPTION_NO_SAME_PERMISSIONS },
+	{ "nuke",		no_argument,	   NULL, OPTION_NUKE },
 	{ "null",		no_argument,	   NULL, OPTION_NULL },
 	{ "one-file-system",	no_argument,	   NULL, OPTION_ONE_FILE_SYSTEM },
 	{ "preserve-permissions", no_argument,     NULL, 'p' },
@@ -242,6 +260,7 @@ main(int argc, char **argv)
 	char			 cachedir[PATH_MAX + 1];
 	char			*homedir;
 	char			*conffile;
+	const char		*missingkey;
 
 	/*
 	 * Use a pointer for consistency, but stack-allocated storage
@@ -332,6 +351,9 @@ main(int argc, char **argv)
 	 */
 	while ((opt = bsdtar_getopt(bsdtar, tar_opts, &option)) != -1) {
 		switch (opt) {
+		case OPTION_AGGRESSIVE_NETWORKING: /* tarsnap */
+			tarsnap_opt_aggressive_networking = 1;
+			break;
 		case 'B': /* GNU tar */
 			/* libarchive doesn't need this; just ignore it. */
 			break;
@@ -376,6 +398,9 @@ main(int argc, char **argv)
 			long_help(bsdtar);
 			exit(0);
 			break;
+		case OPTION_HUMANIZE_NUMBERS: /* tarsnap */
+			tarsnap_opt_humanize_numbers = 1;
+			break;
 		case 'I': /* GNU tar */
 			/*
 			 * TODO: Allow 'names' to come from an archive,
@@ -426,6 +451,12 @@ main(int argc, char **argv)
 		case 'm': /* SUSv2 */
 			bsdtar->extract_flags &= ~ARCHIVE_EXTRACT_TIME;
 			break;
+		case OPTION_MAXBW: /* tarsnap */
+			if (humansize_parse(optarg, &tarsnap_opt_maxbytesout))
+				bsdtar_errc(bsdtar, 1, 0,
+				    "Cannot parse bandwidth limit: %s",
+				    optarg);
+			break;
 		case 'n': /* GNU tar */
 			bsdtar->option_no_subdirs = 1;
 			break;
@@ -467,6 +498,9 @@ main(int argc, char **argv)
 		case OPTION_NODUMP: /* star */
 			bsdtar->option_honor_nodump = 1;
 			break;
+		case OPTION_NOISY_WARNINGS: /* tarsnap */
+			tarsnap_opt_noisy_warnings = 1;
+			break;
 		case OPTION_NO_SAME_OWNER: /* GNU tar */
 			bsdtar->extract_flags &= ~ARCHIVE_EXTRACT_OWNER;
 			break;
@@ -476,6 +510,9 @@ main(int argc, char **argv)
 			bsdtar->extract_flags &= ~ARCHIVE_EXTRACT_XATTR;
 			bsdtar->extract_flags &= ~ARCHIVE_EXTRACT_FFLAGS;
 			break;
+		case OPTION_NUKE: /* tarsnap */
+			set_mode(bsdtar, opt, "--nuke");
+			break;
 		case OPTION_NULL: /* GNU tar */
 			bsdtar->option_null++;
 			break;
@@ -608,7 +645,8 @@ main(int argc, char **argv)
 	if ((bsdtar->tapename == NULL) &&
 	    (bsdtar->mode != OPTION_PRINT_STATS &&
 	     bsdtar->mode != OPTION_LIST_ARCHIVES &&
-	     bsdtar->mode != OPTION_FSCK))
+	     bsdtar->mode != OPTION_FSCK &&
+	     bsdtar->mode != OPTION_NUKE))
 		bsdtar_errc(bsdtar, 1, 0,
 		    "Archive name must be specified");
 	if ((bsdtar->cachedir == NULL) &&
@@ -622,7 +660,10 @@ main(int argc, char **argv)
 		bsdtar_errc(bsdtar, 1, 0,
 		    "Keys must be provided via --keyfile option");
 
-	/* The -f option doesn't make sense for --list-archives and --fsck. */
+	/*
+	 * The -f option doesn't make sense for --list-archives, --fsck, or
+	 * --nuke.
+	 */
 	if ((bsdtar->tapename != NULL) &&
 	    (bsdtar->mode != OPTION_PRINT_STATS))
 		only_mode(bsdtar, "-f", "cxtdr");
@@ -641,7 +682,11 @@ main(int argc, char **argv)
 	if (bsdtar->option_null)
 		only_mode(bsdtar, "--null", "cxt");
 
-	/* Check boolean options only permitted in certain modes. */
+	/* Check options only permitted in certain modes. */
+	if (tarsnap_opt_aggressive_networking)
+		only_mode(bsdtar, "--aggressive-networking", "c");
+	if (tarsnap_opt_maxbytesout != (uint64_t)(-1))
+		only_mode(bsdtar, "--maxbw", "c");
 	if (bsdtar->option_dont_traverse_mounts)
 		only_mode(bsdtar, "--one-file-system", "c");
 	if (bsdtar->option_fast_read)
@@ -690,6 +735,39 @@ main(int argc, char **argv)
 		bsdtar->cachedir = cachedir;
 	}
 
+	/* Make sure we have whatever keys we're going to need. */
+	missingkey = NULL;
+	switch (bsdtar->mode) {
+	case 'c':
+		missingkey = crypto_keys_missing(CRYPTO_KEYMASK_WRITE);
+		break;
+	case 'd':
+	case OPTION_FSCK:
+		missingkey = crypto_keys_missing(CRYPTO_KEYMASK_READ |
+		    CRYPTO_KEYMASK_AUTH_DELETE);
+		break;
+	case OPTION_NUKE:
+		missingkey = crypto_keys_missing(CRYPTO_KEYMASK_AUTH_DELETE);
+		break;
+	case OPTION_PRINT_STATS:
+		/* We don't need keys for printing global stats. */
+		if (bsdtar->tapename == NULL)
+			break;
+
+		/* FALLTHROUGH */
+	case OPTION_LIST_ARCHIVES:
+	case 'r':
+	case 't':
+	case 'x':
+		missingkey = crypto_keys_missing(CRYPTO_KEYMASK_READ);
+		break;
+	}
+	if (missingkey != NULL)
+		bsdtar_errc(bsdtar, 1, 0,
+		    "The %s key is required for %s but is not available",
+		    missingkey, bsdtar->modestr);
+
+	/* Perform the requested operation. */
 	switch(bsdtar->mode) {
 	case 'c':
 		tarsnap_mode_c(bsdtar);
@@ -706,6 +784,9 @@ main(int argc, char **argv)
 	case OPTION_LIST_ARCHIVES:
 		tarsnap_mode_list_archives(bsdtar);
 		break;
+	case OPTION_NUKE:
+		tarsnap_mode_nuke(bsdtar);
+		break;
 	case 'r':
 		tarsnap_mode_r(bsdtar);
 		break;
@@ -1001,8 +1082,8 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 	char * conf_arg;
 	size_t optlen;
 
-	/* Ignore comments. */
-	if (line[0] == '#')
+	/* Ignore comments and blank lines. */
+	if ((line[0] == '#') || (line[0] == '\0'))
 		return (0);
 
 	/* Duplicate line. */
@@ -1023,7 +1104,7 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 
 		/*
 		 * If the line is whitespace-terminated, there might not be
-		 * an option here after all.
+		 * an argument here after all.
 		 */
 		if (conf_arg[0] == '\0')
 			conf_arg = NULL;
@@ -1032,7 +1113,9 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 		conf_arg = NULL;
 	}
 
-	if (strcmp(conf_opt, "cachedir") == 0) {
+	if (strcmp(conf_opt, "aggressive-networking") == 0) {
+		tarsnap_opt_aggressive_networking = 1;
+	} else if (strcmp(conf_opt, "cachedir") == 0) {
 		if (conf_arg == NULL)
 			bsdtar_errc(bsdtar, 1, 0,
 			    "Argument required for "
@@ -1066,6 +1149,9 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 			load_keys(bsdtar, conf_arg);
 			bsdtar->have_keys = 1;
 		}
+	} else if (strcmp(conf_opt, "lowmem") == 0) {
+		if ((bsdtar->mode == 'c') && (bsdtar->cachecrunch == 0))
+			bsdtar->cachecrunch = 1;
 	} else if (strcmp(conf_opt, "nodump") == 0) {
 		if (bsdtar->mode == 'c')
 			bsdtar->option_honor_nodump = 1;
@@ -1091,9 +1177,13 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 	} else if (strcmp(conf_opt, "totals") == 0) {
 		if ((bsdtar->mode == 'c') && (bsdtar->option_totals == 0))
 			bsdtar->option_totals++;
+	} else if (strcmp(conf_opt, "verylowmem") == 0) {
+		if ((bsdtar->mode == 'c') && (bsdtar->cachecrunch == 0))
+			bsdtar->cachecrunch = 2;
 	} else {
 		bsdtar_errc(bsdtar, 1, 0,
-		    "Unrecognized configuration file option: %s", conf_opt);
+		    "Unrecognized configuration file option: \"%s\"",
+		    conf_opt);
 	}
 
 	/* Free memory allocated by strdup. */
